/*-------------------------------------------------------------------*/
/* PNNGBB2.C          Pasi Frnti & Olli Virmajoki                   */
/*                                                                   */
/* Fast PNN method / Global Optimization                             */ 
/*   Branch and Bound / Variant bb2                                  */
/*   Piecewise optimization                                          */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "PNN-GBB2"
#define VersionNumber  "Version 0.12"
#define LastUpdated    "22.9.2004"

/*-------------------------------------------------------------------*/

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <values.h>

#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "sort.h"
#include "solution.h"
#include "pnnfast.h"


/*---------------------------  M i s c	-----------------------------*/


/*#define  STATISTICS 1*/
#ifdef STATISTICS
#define STATCODE(a)a
#else 
#define STATCODE(a)
#endif

#define  round(a)		((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }

static llong cfsol=0LL;
static llong csol=0LL;

/*===================================================================*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/* ========================  OPT  ================================== */


static void CreateDuplicateState(TRAININGSET*   TS,
                                 CODEBOOK*      CB,
                                 PARTITIONING*  P,
                                 CODEBOOK*      CB2,
                                 PARTITIONING*  P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateState(CODEBOOK*     CB2,
                               PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}

static void CreateDuplicateSolution(TRAININGSET*    TS,
                                    CODEBOOK*       CB,
                                    PARTITIONING*   P,
                                    CODEBOOK*       CB2,
                                    PARTITIONING*   P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateSolution(CODEBOOK*    CB2,
                                  PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}


static void Depthsearch(TRAININGSET*  TS,
                        CODEBOOK*     CB,
                        PARTITIONING* P,
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        PARTITIONING* POpt,
                        double*       ErrorOpt,
                        int           a,
                        int           b)
{
  int i, j;
  CODEBOOK     CB2;
  PARTITIONING P2;
  double Error;

  if(b == BookSize(CB))
  {
      a++;
      b=a+1;
  }
  for( i=a; i<FinalSize; i++ )
    {
      for( j= (i==a?b:i+1); j<BookSize(CB); j++ )
      {
      CreateDuplicateSolution(TS, CB, P, &CB2, &P2);

      /* Merge */
      JoinPartitions(TS, &P2, i, j);
      PartitionCentroid(&P2, i, &Node(&CB2, i));
      CopyNode(&Node(&CB2, BookSize(&CB2)-1), &Node(&CB2, j), VectorSize(&CB2));
      BookSize(&CB2)--;
      csol+=1;
      if( BookSize(&CB2) == FinalSize )cfsol+=1;

      Error = AverageErrorCBFast(TS, &CB2, &P2, MSE);
      if( Error < *ErrorOpt )
        {
        if( BookSize(&CB2) == FinalSize )
          {
          *ErrorOpt = Error;
          CopyCodebook(&CB2, CBOpt);
          CopyPartitioning(&P2, POpt);
          }
        else
          {
          Depthsearch(TS, &CB2, &P2, FinalSize, CBOpt, POpt, ErrorOpt, i, j);
          }
        }

      FreeDuplicateSolution(&CB2, &P2);
      }
    }
}


/* ================================================================= */


static void Depthsearch1(TRAININGSET* TS,
                        CODEBOOK*     CB,
                        PARTITIONING* P,
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        PARTITIONING* POpt,
                        double*       Error,
                        double*       ErrorOpt,
                        int           a,
                        int           b)
{
  int i, j;
  CODEBOOK     CB2;
  PARTITIONING P2;
  double Error0=0., EO=0.;
  double* Error2=&Error0;

  if(b == BookSize(CB))
  {
      a++;
      b=a+1;
  }
  for( i=a; i<FinalSize; i++ )
    {
      for( j= ( i==a ? b : i+1); j<BookSize(CB); j++ )
      {
      CreateDuplicateState(TS, CB, P, &CB2, &P2);
      *Error2=*Error;

      *Error2 = *Error2 +
        (((double)VectorFreq(&CB2,i) * (double)VectorFreq(&CB2,j)) /
        ((double)VectorFreq(&CB2,i) + (double)VectorFreq(&CB2,j))) *
        (double)(VectorDist(Vector(&CB2,i), Vector(&CB2,j), 
          VectorSize(&CB2)));

      /* Merge */
      JoinPartitions(TS, &P2, i, j);
      PartitionCentroid(&P2, i, &Node(&CB2, i));
      CopyNode(&Node(&CB2, BookSize(&CB2)-1), &Node(&CB2, j), VectorSize(&CB2));
      BookSize(&CB2)--;
      csol+=1;
      if( BookSize(&CB2) == FinalSize )cfsol+=1;

      EO = AverageErrorCBFast(TS, &CB2, &P2, MSE);
      if( *Error2/((double)TotalFreq(TS)*(double)VectorSize(TS)) < *ErrorOpt )
        {
        if( BookSize(&CB2) == FinalSize )
          {
          *ErrorOpt = *Error2/((double)TotalFreq(TS)*(double)VectorSize(TS));
          printf("\nOK EO %f %f\n",EO, *ErrorOpt);
          CopyCodebook(&CB2, CBOpt);
          CopyPartitioning(&P2, POpt);
          }
        else
          {
          Depthsearch1(TS, &CB2, &P2, FinalSize, CBOpt, POpt, Error2, ErrorOpt, i, j);
          }
        }
      FreeDuplicateState(&CB2, &P2);
      }
    }
}


static void CentroidOfTwoVectors(CODEBOOK* CB, BOOKNODE* cent, int p1, int p2)
{
  int k;
  int freqsum;

  freqsum = VectorFreq(CB,p1) + VectorFreq(CB,p2);

  for( k=0; k<VectorSize(CB); k++)
    {
	if( freqsum > 0 )
	cent->vector[k] = round( (double)(VectorFreq(CB, p1)*VectorScalar(CB, p1, k) +
                                      VectorFreq(CB, p2)*VectorScalar(CB, p2, k))
                             / (double)freqsum );
	else cent->vector[k] = 0;
    }
  cent->freq = freqsum;
}


static void Depthsearch2(CODEBOOK*    CB,
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        double*       Error,
                        double*       ErrorOpt,
                        double        Tfvs,
                        int           a,
                        int           b)
{
  int i, j;
  CODEBOOK CB2;
  double Error0=0.;
  double* Error2=&Error0;

  if(b == BookSize(CB))
  {
      a++;
      b=a+1;
  }
  for( i=a; i<FinalSize; i++ )
    {
      for( j= ( i==a ? b : i+1); j<BookSize(CB); j++ )
      {
      CreateNewCodebook(&CB2, BookSize(CB), CB);
      CopyCodebook(CB, &CB2);
      *Error2=*Error;

      *Error2 = *Error2 + (double)(VectorDist(Vector(&CB2,i), Vector(&CB2,j), 
          VectorSize(&CB2)))*
        ((double)VectorFreq(&CB2,i) * (double)VectorFreq(&CB2,j)) /
        ((double)VectorFreq(&CB2,i) + (double)VectorFreq(&CB2,j));

      /* Merge */
      CentroidOfTwoVectors(&CB2, &Node(&CB2, i), i, j); 
      CopyNode(&Node(&CB2, BookSize(&CB2)-1), &Node(&CB2, j), VectorSize(&CB2));
      BookSize(&CB2)--;

      csol+=1;
      if( BookSize(&CB2) == FinalSize )cfsol+=1;

      if( *Error2/Tfvs < *ErrorOpt )
        {
        if( BookSize(&CB2) == FinalSize )
          {
            //printf("\nOK");
            *ErrorOpt = *Error2/Tfvs;
            CopyCodebook(&CB2, CBOpt); 
          }
        else
          {
            Depthsearch2(&CB2, FinalSize, CBOpt, Error2, ErrorOpt, Tfvs, i, j);
          }
        }
      FreeCodebook(&CB2);
      }
    }
}


/* ================================================================= */


double PNNGBb2(TRAININGSET*	    TS,
			   CODEBOOK*		CB,
			   PARTITIONING*	P,
               int              FinalSize,
               int              StepSize,
               int              Quiet,
               long             watch,
               int              Pds,
               int              Mps,
               int              Lazy)
{
  CODEBOOK CBOpt/*, CBOpt2*/;
  PARTITIONING POpt/*, POpt2*/;
  int NewSize;
  double ErrorOpt=MAXLLONG, ErrorOpt2=MAXLLONG; 
  //llong EO=MAXLLONG;
  double Error0=0LL;
  double *Error = &Error0;
  int Tfvs = TotalFreq(TS)*VectorSize(TS);

  Pds = 0;
  Mps = 0;
  Lazy = 0;

  if(BookSize(CB) > FinalSize)
  {

  while(BookSize(CB) > FinalSize)
  {
  NewSize = BookSize(CB) - StepSize;
  if(NewSize < FinalSize)
      NewSize = FinalSize;

  CreateDuplicateSolution(TS, CB, P, &CBOpt, &POpt);
  PNNFast(TS, &CBOpt, &POpt, NewSize, 0, watch, Pds, Mps, Lazy);
  ErrorOpt = AverageErrorCBFast(TS, &CBOpt, &POpt, MSE);

  if(Quiet > 1)
  {
  printf("1 %f %f ",ErrorOpt, (double)*Error);
  PrintTime(watch);
  }

  //Depthsearch(TS, CB, P, NewSize, &CBOpt, &POpt, &ErrorOpt, 0, 1);
  //ErrorOpt=MAXLLONG;
  Depthsearch1(TS, CB, P, NewSize, &CBOpt, &POpt, Error, &ErrorOpt, 0, 1);
  //Depthsearch2(CB, NewSize, &CBOpt, Error, &ErrorOpt, Tfvs, 0, 1);

  *Error = ErrorOpt*Tfvs;
  if(Quiet > 1)
  {
  printf("2 %f %f ",ErrorOpt, (double)*Error); 
  PrintTime(watch);
  }

  ErrorOpt2 = AverageErrorCBFast(TS, &CBOpt, &POpt, MSE);
  *Error = ErrorOpt2*Tfvs;

  if(Quiet > 1)
  {
  printf("3 %f %f ",ErrorOpt2, (double)*Error );
  PrintTime(watch);
  }

  CopyCodebook(&CBOpt, CB);
  CopyPartitioning(&POpt, P);

  FreeDuplicateSolution(&CBOpt, &POpt);

  printf("Booksize: %d \n\n",BookSize(CB));
  }

  
  }
  else
  {
    ErrorOpt = AverageErrorCBFast(TS, CB, P, MSE);
  }

  printf("Final solutions: %7.0f All solutions: %7.0f \n",(double)cfsol,(double)csol);
  return (ErrorOpt);
}

