/*-------------------------------------------------------------------*/
/* PNNGBB3.C          Pasi Frnti & Olli Virmajoki                   */
/*                                                                   */
/* Fast PNN method / Global Optimization                             */ 
/*   Branch and Bound / Variant bb3                                  */
/*   Look-ahead optimization                                         */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "PNN-GBB3"
#define VersionNumber  "Version 0.9"
#define LastUpdated    "9.12.2001" 
/*-------------------------------------------------------------------*/

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <values.h>

#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "sort.h"
#include "solution.h"
#include "pnnfast.h"


/*---------------------------  M i s c	-----------------------------*/


/*#define  STATISTICS 1*/
#ifdef STATISTICS
#define STATCODE(a)a
#else 
#define STATCODE(a)
#endif

#define  round(a)		((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }

static llong cfsol=0LL;
static llong csol=0LL;

/*===================================================================*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/*-------------------------------------------------------------------*/


/* ========================  OPT  ================================== */


static void CreateDuplicateSolution(TRAININGSET*    TS,
                                    CODEBOOK*       CB,
                                    PARTITIONING*   P,
                                    CODEBOOK*       CB2,
                                    PARTITIONING*   P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateSolution(CODEBOOK*    CB2,
                                  PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}


static void Depthsearch(TRAININGSET*  TS,
                        CODEBOOK*     CB,
                        PARTITIONING* P,
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        PARTITIONING* POpt,
                        double*       ErrorOpt,
                        CODEBOOK*     CBI,
                        PARTITIONING* PI,
                        CODEBOOK*     CBOptI,
                        PARTITIONING* POptI,
                        int           NewSizeI,
                        int           a,
                        int           b)
{
  int i, j;
  CODEBOOK     CB2;
  PARTITIONING P2;
  double Error;

  if(b==BookSize(CB))
  {
      a++;
      b=a+1;
  }
  for( i=a; i<FinalSize; i++ )
    {
      for( j=(i==a?b:i+1); j<BookSize(CB); j++ )
      {
      CreateDuplicateSolution(TS, CB, P, &CB2, &P2);

      /* Merge */
      JoinPartitions(TS, &P2, i, j);
      PartitionCentroid(&P2, i, &Node(&CB2, i));
      CopyNode(&Node(&CB2, BookSize(&CB2)-1), &Node(&CB2, j), VectorSize(&CB2));
      BookSize(&CB2)--;
      csol+=1;
      if( BookSize(&CB2) == FinalSize ) cfsol+=1;

      if(BookSize(&CB2)==NewSizeI)
            {
                CopyCodebook(&CB2, CBI);
                CopyPartitioning(&P2, PI);
            }

      Error = AverageErrorCBFast(TS, &CB2, &P2, MSE);
      if( Error < *ErrorOpt )
        {
        if( BookSize(&CB2) == FinalSize )
          {
          *ErrorOpt = Error;
          CopyCodebook(&CB2, CBOpt);
          CopyPartitioning(&P2, POpt);
                   
          CopyCodebook(CBI, CBOptI);
          CopyPartitioning(PI, POptI);

          }
        else
          {
          Depthsearch(TS, &CB2, &P2, FinalSize, CBOpt, POpt, ErrorOpt, CBI, PI, CBOptI, POptI, NewSizeI, i, j);
          }
        }
      FreeDuplicateSolution(&CB2, &P2);

      }
    }
}


/* ================================================================= */


static void CentroidOfTwoVectors(CODEBOOK* CB, BOOKNODE* cent, int p1, int p2)
{
  int k;
  int freqsum;

  freqsum = VectorFreq(CB,p1) + VectorFreq(CB,p2);

  for( k=0; k<VectorSize(CB); k++)
    {
	if( freqsum > 0 )
	cent->vector[k] = round( (double)(VectorFreq(CB, p1)*VectorScalar(CB, p1, k) +
                                      VectorFreq(CB, p2)*VectorScalar(CB, p2, k))
                             / (double)freqsum );
	else cent->vector[k] = 0;
    }
  cent->freq = freqsum;
}


static void Depthsearch2(CODEBOOK*    CB,
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        double*       Error,
                        double*       ErrorOpt,
                        CODEBOOK*     CBI,
                        CODEBOOK*     CBOptI,
                        int           NewSizeI,
                        double        Tfvs,
                        int           a,
                        int           b)
{
  int i, j;
  CODEBOOK CB2;
  double Error0=0.;
  double* Error2=&Error0;

  if(b == BookSize(CB))
  {
      a++;
      b=a+1;
  }
  for( i=a; i<FinalSize; i++ )
    {
      for( j= ( i==a ? b : i+1); j<BookSize(CB); j++ )
      {
      CreateNewCodebook(&CB2, BookSize(CB), CB);
      CopyCodebook(CB, &CB2);
      *Error2=*Error;

      *Error2 = *Error2 + (double)(VectorDist(Vector(&CB2,i), Vector(&CB2,j), 
          VectorSize(&CB2)))*
        ((double)VectorFreq(&CB2,i) * (double)VectorFreq(&CB2,j)) /
        ((double)VectorFreq(&CB2,i) + (double)VectorFreq(&CB2,j));

      /* Merge */
      CentroidOfTwoVectors(&CB2, &Node(&CB2, i), i, j); 
      CopyNode(&Node(&CB2, BookSize(&CB2)-1), &Node(&CB2, j), VectorSize(&CB2));
      BookSize(&CB2)--;

      csol+=1;
      if( BookSize(&CB2) == FinalSize )cfsol+=1;

      if(BookSize(&CB2)==NewSizeI)
      {
                CopyCodebook(&CB2, CBI);
      }

      if( *Error2/Tfvs < *ErrorOpt )
        {
        if( BookSize(&CB2) == FinalSize )
          {
            *ErrorOpt = *Error2/Tfvs;
            CopyCodebook(&CB2, CBOpt);
            CopyCodebook(CBI, CBOptI);
          }
        else
          {
            Depthsearch2(&CB2, FinalSize, CBOpt, Error2, ErrorOpt, CBI, CBOptI, NewSizeI, Tfvs, i, j);
          }
        }
      FreeCodebook(&CB2);
      }
    }
}


/* ================================================================= */


double PNNGBb3(TRAININGSET*	    TS,
			   CODEBOOK*		CB,
			   PARTITIONING*	P,
               int              FinalSize,
               int              StepSize,
               int              Quiet,
               long             watch,
               int              Pds,
               int              Mps,
               int              Lazy)
{
  CODEBOOK CBOpt,CBOptI, CBI;
  PARTITIONING POpt, POptI, PI;
  int NewSize, NewSizeI;
  double ErrorOpt;
  //double Error0=.0;
  //double *Error = &Error0;
  //double Tfvs = (double)(TotalFreq(TS)*VectorSize(TS));

  Pds = 0;
  Mps = 0;
  Lazy = 0;

  if(BookSize(CB) > FinalSize)
  {

  while(BookSize(CB) > FinalSize)
  {
  NewSize = BookSize(CB) - StepSize;
  if(NewSize < FinalSize)
      NewSize = FinalSize;
  NewSizeI = BookSize(CB)-1; 

  CreateDuplicateSolution(TS, CB, P, &CBOpt, &POpt);
  CreateDuplicateSolution(TS, CB, P, &CBI, &PI);
  CreateDuplicateSolution(TS, CB, P, &CBOptI, &POptI);

  PNNFast(TS, &CBOptI, &POptI, NewSizeI, Quiet, watch, Pds, Mps, Lazy);
  PNNFast(TS, &CBOpt, &POpt, NewSize, Quiet, watch, Pds, Mps, Lazy);
  (ErrorOpt) = AverageErrorCBFast(TS, &CBOpt, &POpt, MSE);

  Depthsearch(TS, CB, P, NewSize, &CBOpt, &POpt, &ErrorOpt, &CBI, &PI, &CBOptI, &POptI, NewSizeI, 0, 1);
  //Depthsearch2(CB, NewSize, &CBOpt, Error, &ErrorOpt, &CBI, &CBOptI, NewSizeI, Tfvs, 0, 1);
  
  CopyCodebook(&CBOptI, CB);
  CopyPartitioning(&POptI, P);

  FreeDuplicateSolution(&CBOpt, &POpt);
  FreeDuplicateSolution(&CBI, &PI);
  FreeDuplicateSolution(&CBOptI, &POptI);

  printf("Booksize: %d \n",BookSize(CB));
  }
  
  }
  else
  {
  (ErrorOpt) = AverageErrorCBFast(TS, CB, P, MSE);
  }

  printf("Final solutions: %7.0f All solutions: %7.0f \n",(double)cfsol,(double)csol);
  return (ErrorOpt);
}

